**************************** 32-bit divide ********************
*                        *** Pseudocode ***
;  WHILE divisor < dividend DO
;    scale factor := scale factor + 1
;    divisor := divisor * 2
;
;  quotient := 0
;  REPEAT
;    WHILE dividend > divisor DO
;      dividend := dividend - divisor
;      quotient := quotient OR 2^scale factor
;    divisor := divisor / 2
;    scale factor := scale factor - 1
;  UNTIL (scale factor < 0) 
;------------------------------------------------------------------
; stack:   rtn address +0
;          dvsr        +4
;          dvd         +8
;......................... register usage .......................
;          d0   dvsr
;          d1   dvd
;          d2   scale factor
;          d3   sign count (count of negative operands)
;          d4   quotient 
;          d5   
;          d6   
*-----------------------------------------------------------------
savlist REG     d1-d4             
savbias EQU     4*17
div32:
;       Since this routine has no spawn, it's ok to save the registers on
;       the stack without anyones' permission or knowledge
        movem.l savlist,-savbias(sp)    ****
        
;       count the signs as the numbers are loaded
;       will correct the sign after computing the quotient
        clr.l   d3
        move.l  4(sp),d0
        beq     div_zero                eror exit
        bpl     pos1
        neg.l   d0
        moveq   #1,d3                   count the negative sign
pos1:        
        move.l  8(sp),d1
        bpl     pos2
        neg.l   d1
        addq.l  #1,d3
pos2:        
        clr.l   d4
        clr.l   d2

;       ASSERT: there is a positive number in d0 and also in d1.  
;               d0 is non-zero.        
;               divide d1 by d0
.loop:
        cmp.l   d0,d1
        bls     .skip
        addq.b  #1,d2                   d2 = scale factor
        add.l   d0,d0                   d0 := 2 * d0
        bra     .loop
;       ...
        
;       Assert:  dvsr = dvsr * 2^scale factor
.skip:        
        moveq   #0,d4                   quotient = 0  (.long) 
.loop2:        
        cmp.l   d0,d1
        bcs     .skip2                  branch if carry
        sub.l   d0,d1                   d1 := d1 - d0
        bset    d2,d4                   d4 := d4 OR 2^d2
.skip2:        
        lsr.l   #1,d0                   d0 := d0 / 2
        dbeq    d2,.loop2
;       ...
;       quotient for unsigned number is in d4        

;       save the remainder so modulus can be computed
        move.l  d1,remainder        
        
;       fiddle the sign. d3 contains count of negative numbers at entry
        tst     d3
        beq     .result_plus            ct = 0, exit
        goeql   d3,#2,.result_plus       
        beq     .result_plus            ct = 2, exit
        neg.l   d4                      one or two negative inputs => negative output
        neg.l   remainder               fix sign for modulus, too
.result_plus:           
        move.l  d4,d0                   result of FUNCTION
        movem.l -savbias(sp),savlist
        return   8                      return to caller
;       ...        
div_zero:
        pea     err_msg
        move    #9,-(sp)                cconws
        trap    #1                      GEMDOS
        addq.l  #6,sp
        
        clr     -(sp)                   pterm0
        trap    #1        
;       ...        
;============================================================
        SECTION DATA
        
err_msg DC.B    13,10,' Attempt to divide by zero',13,10
        DC.B          ' in 32-bit divide routine',13,10,0        
        SECTION BSS
remainder   DS.L        1            
        SECTION TEXT
*************************************************
*
*  Fill the registers with a test pattern  3/9/89
*    See if the successor routines are well behaved
*    Call:    bsr fill_reg
************************************************
fill_reg:
        move.l  #$2222220E,d0   mark is a bunch of hex 2s followed by
                                * a register number [0..15]
        * $0E = 15 decimal
        move    #14,d1          load 15 registers
.loop
        move.l  d0,-(sp)        push marker on stack
        sub.l   #1,d0           generate next marker
        dbra    d1,.loop
        movem.l (sp)+,d0-d7/a0-a6
        rts

*********************** multiply two 32-bit numbers *****************
*============== register usage =============
;       d0 current sum
;       d1 op1
;       d2 op2
;       d3 repeat count         
;====================== pseudocode ============================
; FOR bit := 0 TO 30 DO
;   IF lsb(op1) = 1
;     THEN sum := sum + op2
;   op1 := op1 / 2                    
;   op2 := op2 * 2                    
;    (if a 1-bit is shifted off, it represents an overflow.  
;     no test are made for overflow.)        
;--------------------------- code ------------------------------
myreg   REG     d1-d3
regbias EQU     4*17

mpy32:  
        movem.l myreg,-regbias(sp)
        clr.l   d0                      current sum
        abs     4(sp),d1
        abs     8(sp),d2
;       minimize the number of leading zero bits in the operand in d1.
;       this reduces the number of iterations we have to perform, and 
;       thus increases the speed of evaluation.        
        min.l   d1,d2                   smaller operand to d1
        move    #32,d3                  repeat count
        gogtr.l d1,#16777215,loop415    2^24 -1
        move    #25,d3
        gogtr.l d1,#65535,loop415       2^16 - 1
        move    #17,d3
        
loop415:
        btst    #0,d1
        beq     .no_add
        add.l   d2,d0
.no_add:
        asr.l   #1,d1
        asl.l   #1,d2
        dbra    d3,loop415
;       ...        
;     *** result is in d0 ***
;       now correct the sign
        move.l  4(sp),d1                one op
        bpl     is_pos
        neg.l   d0
is_pos  move.l  8(sp),d1                other op
        bpl     pos_again
        neg.l   d0
pos_again:        
        movem.l -regbias(sp),myreg
;       strip params and return
        move.l  (sp),8(sp)
        add.l   #8,sp
        rts
*------------------------------ end multiply ---------------------
